﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;

/*
	SkyboxColor is plugin to change:
		Skybox/Panoramic
		Skybox/Cubemap
		Unlit/Texture (Skybox/6 Sided)
	material in CUA (.assetbundle)

	Based on UnityAssetVamifier code from NoStage3
	Made by DemonLord
	License: Creative Commons with Attribution (CC BY)
*/

namespace DLPlugin {
	public class SkyboxColor : MVRScript {
		private readonly string VERSION_NUMBER = "1.00";

		//Valid shader names
		private readonly List<string> UNITY_SHADER_NAMES = new List<string>(new string[] {
			"Skybox/Panoramic",
			"Skybox/Cubemap",
			"Unlit/Texture",
			"Skybox/6 Sided"
		});
		//material names for "Skybox/6 Sided" case
		private readonly List<string> SIDE_NAMES = new List<string>(new string[] {
			"back (instance)",
			"bottom (instance)",
			"left (instance)",
			"right (instance)",
			"front (instance)",
			"top (instance)",
			"up (instance)",
			"down (instance)"
		});

		protected UIDynamicSlider exposureSlider;
		protected UIDynamicColorPicker tintColorPicker;
		protected UIDynamicToggle showMessagesToggle;

		protected JSONStorableFloat jExposureFloat;
		protected JSONStorableColor jTintColor;
		protected JSONStorableBool showMessagesBool;

		protected List<MatPackage> origMaterials = new List<MatPackage>();
		protected List<MatPackage> newMaterials = new List<MatPackage>();
		protected List<Renderer> allRenderers = new List<Renderer>();
		protected CustomUnityAssetLoader customUnityAssetLoader;

		public struct MatPackage {
			public int rendererIndex;
			public int materialIndex;
			public Material material;
			public bool isConverted;

			public MatPackage(int rIndex, int mIndex, Material mat, bool converted) {
				this.rendererIndex = rIndex;
				this.materialIndex = mIndex;
				this.material = mat;
				this.isConverted = converted;
			}
		}

		public override void Init() {
			try {

				// build/load/populate material controls for VAM UI
				BuildUIControls();

				AddVersionNumberToUI();

				// get the loader component
				customUnityAssetLoader = containingAtom.GetComponentInChildren<CustomUnityAssetLoader>();
#if VAM_GT_1_20
				if (customUnityAssetLoader != null) {
					customUnityAssetLoader.RegisterAssetClearedCallback(AssetCleared);
					customUnityAssetLoader.RegisterAssetLoadedCallback(AssetLoaded);
					if (customUnityAssetLoader.isAssetLoaded) {
						// already loaded - sync up
						AssetLoaded();
					}
				}
#else
				// since there is no "asset bundle loaded" callback yet, we have to wait
				StartCoroutine(WaitForAssetBundle());
#endif

			}
			catch (Exception e) {
				SuperController.LogError("SkyboxColor: Exception caught: " + e);
			}
		}


		protected void BuildUIControls() {
			// Exposure 
			jExposureFloat = new JSONStorableFloat("Exposure", 0.5f, SetExposure, 0f, 5f, true);
			RegisterFloat(jExposureFloat);
			exposureSlider = CreateSlider(jExposureFloat);

			//Tint Color
			HSVColor tintColorHSVC = HSVColorPicker.RGBToHSV(1f, 1f, 1f);
			jTintColor = new JSONStorableColor("TintColor", tintColorHSVC, SetTintColor);
			RegisterColor(jTintColor);
			tintColorPicker = CreateColorPicker(jTintColor, true);

			// showMessagesToggle
			showMessagesBool = new JSONStorableBool("Show Messages", true);
			RegisterBool(showMessagesBool);
			showMessagesToggle = CreateToggle(showMessagesBool, true);
		}


		protected void AddVersionNumberToUI() {
			JSONStorableString versionString = new JSONStorableString("pluginVersion", "\nSkyboxColor "+ VERSION_NUMBER + "\nby DemonLord (CC BY)");
			UIDynamicTextField textFieldVersion = CreateTextField(versionString, true);
			textFieldVersion.height = 120;
		}


		// determine if the material should be converted or not
		protected bool IsValidMaterial(Material mat) {
			if (!UNITY_SHADER_NAMES.Contains(mat.shader.name))
				return false;
			return true;
		}
		
		// check material name as one of 6 Sided names
		protected bool IsValidSide(Material mat) {
			if (!SIDE_NAMES.Contains(mat.name.ToLower()))
				return false;
			return true;
		}

		protected void AssetCleared() {
			// don't need to reset materials because the asset would have been destroyed
			allRenderers.Clear();
			origMaterials.Clear();
			newMaterials.Clear();
		}

		protected void AssetLoaded() {
			containingAtom.GetComponentsInChildren<Renderer>(false, allRenderers);
			ConvertMaterial();
		}

		// wait for the asset bundle to be loaded by checking for renderers every 0.5s (a bit hacky...)
		protected IEnumerator WaitForAssetBundle() {

			while (allRenderers.Count < 1) {
				containingAtom.GetComponentsInChildren<Renderer>(false, allRenderers);
				yield return new WaitForSeconds(0.5f);
			}

			ConvertMaterial();
		}


		protected void ConvertMaterial() {
			List<Material> invalidMaterials = new List<Material>();

			if (allRenderers.Count == 0) {
				if (showMessagesBool.val) {
					SuperController.LogMessage("SkyboxColor: Make sure you have an .assetbundle loaded and a prop selected from the Asset dropdown. Then reload this plugin.");
				}
#if VAM_GT_1_20
				else {
					SuperController.LogMessage("SkyboxColor: Make sure you have an .assetbundle loaded and a prop selected from the Asset dropdown. Then reload this plugin.", true, false);
				}
#endif
				return;
			}

			// fill newMaterials with copy of the original material or an new material with the Skybox/6 Sided shader
			for (int i = 0; i < allRenderers.Count; i++) {
				for (int j = 0; j < allRenderers[i].materials.Length; j++) {
					Material mat = allRenderers[i].materials[j];
					origMaterials.Add(new MatPackage(i, j, mat, false));

					if (IsValidMaterial(mat)) {
						if (mat.shader.name.Contains("Unlit/Texture") && IsValidSide(mat)){
							newMaterials.Add(new MatPackage(i, j, new Material(Shader.Find("Skybox/6 Sided")), true));
						}else{
							newMaterials.Add(new MatPackage(i, j, mat, true));
						}
					} else {
						newMaterials.Add(new MatPackage(i, j, mat, false));
						invalidMaterials.Add(mat);
					}
				}
			}


			// output invalid/skipped materials
			foreach (Material invalidMaterial in invalidMaterials) {
				if (showMessagesBool.val) {
					SuperController.LogMessage("SkyboxColor: Material skipped: " + invalidMaterial.name + " -- Shader: [" + invalidMaterial.shader.name + "]");
				}
#if VAM_GT_1_20
				else {
					SuperController.LogMessage("SkyboxColor: Material skipped: " + invalidMaterial.name + " -- Shader: [" + invalidMaterial.shader.name + "]", true, false);
				}
#endif
			}

			if (newMaterials.Count == 0) {
				SuperController.LogError("SkyboxColor: No compatible material could be found on this asset. Check message log for skipped materials/shaders.");
				return;
			}


			// replace renders from original to copy materials
			for (int i = 0; i < allRenderers.Count; i++) {
				Material[] matsToAssign = allRenderers[i].materials;
				for (int j = 0; j < allRenderers[i].materials.Length; j++) {
					Material mat = allRenderers[i].materials[j];
					if (IsValidMaterial(mat)) {
						Material vamMat = GetVamMaterialByIndex(i, j);

						// for "Skybox/6 Sided" case
						if (mat.shader.name.Contains("Unlit/Texture") && IsValidSide(mat)){
							Material origMat = GetOrigMaterialByIndex(i, j);
							vamMat.SetTexture("_LeftTex", origMat.GetTexture("_MainTex"));
							vamMat.SetTexture("_RightTex", origMat.GetTexture("_MainTex"));
							vamMat.SetTexture("_FrontTex", origMat.GetTexture("_MainTex"));
							vamMat.SetTexture("_BackTex", origMat.GetTexture("_MainTex"));
							vamMat.SetTexture("_UpTex", origMat.GetTexture("_MainTex"));
							vamMat.SetTexture("_DownTex", origMat.GetTexture("_MainTex"));
						}
						matsToAssign[j] = vamMat;
					}
				}
				// assign new materials to prop renderer as array
				allRenderers[i].materials = matsToAssign;
			}

			// apply user definable values initially
			SetExposure(jExposureFloat);
			SetTintColor(jTintColor);
		}

		protected Material GetVamMaterialByIndex(int rIndex, int mIndex) {
			foreach (MatPackage matPackage in newMaterials) {
				if (matPackage.rendererIndex == rIndex && matPackage.materialIndex == mIndex) {
					return matPackage.material;
				}
			}
			return null;
		}

		void OnDestroy() {
			RestoreOriginalMaterials();
#if VAM_GT_1_20
				if (customUnityAssetLoader != null) {
					customUnityAssetLoader.DeregisterAssetClearedCallback(AssetCleared);
					customUnityAssetLoader.DeregisterAssetLoadedCallback(AssetLoaded);
				}
#endif
		}

		protected Material GetOrigMaterialByIndex(int rIndex, int mIndex) {
			foreach (MatPackage matPackage in origMaterials) {
				if (matPackage.rendererIndex == rIndex && matPackage.materialIndex == mIndex) {
					return matPackage.material;
				}
			}
			return null;
		}

		protected void RestoreOriginalMaterials() {
			for (int i = 0; i < allRenderers.Count; i++) {
				if (allRenderers[i] != null) {   // renderer may have been destroyed already by atom remove
					List<Material> matsToAssign = new List<Material>();
					for (int j = 0; j < allRenderers[i].materials.Length; j++) {
						matsToAssign.Add(GetOrigMaterialByIndex(i, j));
					}
					allRenderers[i].materials = matsToAssign.ToArray();
				}
			}
		}

		protected void SetExposure(JSONStorableFloat jf) {
			foreach (MatPackage matPackage in newMaterials) {
				if (!matPackage.isConverted) continue;

				matPackage.material.SetFloat("_Exposure", jf.val);
			}
		}

		protected void SetTintColor(JSONStorableColor jcolor) {
			foreach (MatPackage matPackage in newMaterials) {
				if (!matPackage.isConverted) continue;
				matPackage.material.SetColor("_Tint", jcolor.colorPicker.currentColor);
			}
		}


	}
}